"""Test PyAudioSourceReader"""
import time
from contextlib import contextmanager
from unittest.mock import patch

from audiostream2py import PyAudioSourceReader, PaStatusFlags, AudioSegment


@contextmanager
def mock_pyaudio_context():
    class MockPyAudio:
        @classmethod
        def return_true(cls, *_a, **_kw):
            return True

        def __getattr__(self, item):
            return self.return_true

    yield MockPyAudio()


def test_stream_callback():
    """Test normal data reads and input overflow"""

    input_device_index = 999  # fake device index
    width = 2
    channels = 1
    frames_per_buffer = 10
    rate = 10

    with patch('time.time') as mock_time, patch(
        'audiostream2py.PyAudioSourceReader._pyaudio'
    ) as mock_source_pyaudio:
        mock_source_pyaudio.return_value = mock_pyaudio_context()
        mock_time.side_effect = list(range(1, 100))
        mock_pcm_bytes = b''.join(
            i.to_bytes(width, 'little') * channels for i in range(frames_per_buffer)
        )
        source = PyAudioSourceReader(
            input_device_index=input_device_index,
            width=width,
            channels=channels,
            frames_per_buffer=frames_per_buffer,
            rate=rate,
        )
        # simulate normal
        for i in range(1, 10):
            source._stream_callback(
                mock_pcm_bytes, frames_per_buffer, None, PaStatusFlags.paNoError
            )
            assert len(source.data) == i
            audio_segment: AudioSegment = source.data[-1]
            assert audio_segment.waveform == mock_pcm_bytes
            assert audio_segment.frame_count == frames_per_buffer
            assert audio_segment.start_date == (i - 1) * 1e6
            assert audio_segment.end_date == i * 1e6

        data_count = len(source.data)
        last_audio_segment: AudioSegment = source.data[-1]
        # simulate input overflow
        input_overflow_time = 10
        for _ in range(input_overflow_time):
            time.time()  # increment mock_time
            source._stream_callback(
                mock_pcm_bytes, frames_per_buffer, None, PaStatusFlags.paInputOverflow
            )
            assert len(source.data) == data_count
            audio_segment: AudioSegment = source.data[-1]
            assert audio_segment == last_audio_segment

        # simulate first normal after input overflow
        source._stream_callback(
            mock_pcm_bytes, frames_per_buffer, None, PaStatusFlags.paNoError
        )
        assert len(source.data) == data_count + 2
        error_filler: AudioSegment = source.data[-2]
        final_audio_segment: AudioSegment = source.data[-1]

        assert error_filler.waveform == (
            PyAudioSourceReader.FILL_VALUE.to_bytes(width, 'little')
            * channels
            * frames_per_buffer
            * input_overflow_time
        )

        assert error_filler.frame_count == frames_per_buffer * input_overflow_time
        assert error_filler.start_date == last_audio_segment.end_date
        assert error_filler.end_date == final_audio_segment.start_date
        assert error_filler.start_date != error_filler.end_date

        assert final_audio_segment.waveform == mock_pcm_bytes
        assert final_audio_segment.frame_count == frames_per_buffer
        assert (
            final_audio_segment.start_date
            == last_audio_segment.start_date + (input_overflow_time + 1) * 1e6
        )
        assert (
            final_audio_segment.end_date
            == last_audio_segment.start_date + (input_overflow_time + 2) * 1e6
        )



"""WavFileSourceReader to read a wav file and produce read output similar to PyAudioSourceReader."""

import wave
from functools import cached_property
from io import BytesIO
from pathlib import Path
from typing import Union

from stream2py import SourceReader
from stream2py.utility.typing_hints import ComparableType

from audiostream2py import AudioSegment, PaStatusFlags
from audiostream2py.reader import AudioBufferReader


def resolve_file(file: Union[str, bytes, Path, BytesIO]) -> Union[str, BytesIO]:
    """Resolve file into format excepted by wave.open()

    :param file: file path or bytes
    :return: str or BytesIO
    """
    if isinstance(file, bytes):
        return BytesIO(file)
    if isinstance(file, Path):
        return str(file)
    return file


def file_to_audio_segment(
    file: Union[str, bytes, Path, BytesIO], start_date=0
) -> AudioSegment:
    """Read wav file into an AudioSegment

    :param file: file path or bytes
    :param start_date: timestamp in microseconds
    :return:
    """
    file = resolve_file(file)
    with wave.open(file, 'rb') as wave_file:
        framerate = wave_file.getframerate()
        nframes = wave_file.getnframes()
        end_date = start_date + nframes * 1e6 / framerate
        audio_data = wave_file.readframes(nframes)

        return AudioSegment(
            start_date=start_date,
            end_date=end_date,
            waveform=audio_data,
            frame_count=nframes,
            status_flags=PaStatusFlags.paNoError,
        )


class WavFileSourceReader(SourceReader):
    buffer_reader_class = AudioBufferReader  # BufferReader specific to AudioSegment

    def __init__(
        self,
        file: Union[str, bytes, Path, BytesIO],
        *,
        frames_per_buffer=1024,
        start_date=0
    ):
        super().__init__()
        self.file = resolve_file(file)
        self.frames_per_buffer = frames_per_buffer
        self.start_date = start_date
        self._fp: wave.Wave_read = wave.open(self.file, 'rb')

        self._index = 0
        self._next_bt = self.start_date

    @property
    def info(self) -> dict:
        return {
            'channels': self._fp.getnchannels(),
            'width': self._fp.getsampwidth(),
            'rate': self._fp.getframerate(),
            'n_frames': self._fp.getnframes(),
            'frames_per_buffer': self.frames_per_buffer,
            'bt': self.start_date,
        }

    def open(self) -> None:
        self._fp.rewind()
        self._index = 0
        self._next_bt = self.start_date

    def close(self) -> None:
        pass

    @cached_property
    def sr(self) -> int:
        return self._fp.getframerate()

    @cached_property
    def bytes_per_frame(self) -> int:
        width = self._fp.getsampwidth()
        n_channels = self._fp.getnchannels()
        return width * n_channels

    def read(self) -> AudioSegment:
        data = self._fp.readframes(self.frames_per_buffer)
        if len(data) == 0:
            return None

        n_frames = int(len(data) / self.bytes_per_frame)
        bt = self._next_bt
        tt = bt + n_frames * 1e6 / self.sr
        self._next_bt = tt
        return AudioSegment(
            start_date=bt,
            end_date=tt,
            waveform=data,
            frame_count=n_frames,
            status_flags=PaStatusFlags.paNoError,
        )

    def key(self, data: AudioSegment) -> ComparableType:
        """AudioSegment is a ComparableType

        :param data: {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}
        :return: AudioSegment
        """
        return data.start_date

    def __del__(self):
        self._fp.close()



"""AudioSegment dataclass for read values from PyAudioSourceReader.  Provides methods to concatenate
multiple AudioSegment objs and slicing an individual AudioSegment obj while maintaining the correct
timestamps and frame counts.
"""

import math
from functools import cached_property
from typing import Union, Sequence, Tuple, Literal, Optional
from dataclasses import dataclass

from audiostream2py.enum import PaStatusFlags


@dataclass(frozen=True)
class AudioSegment:
    start_date: Union[int, float]
    end_date: Union[int, float]
    waveform: bytes
    frame_count: int
    status_flags: PaStatusFlags

    def __post_init__(self):
        if not self.is_empty():
            if self.end_date <= self.start_date:
                raise ValueError('end_date must be higher than start_date')
            if len(self.waveform) == 0:
                raise ValueError(
                    'waveform cannot be empty. Use AudioSegment.empty() to create empty segment'
                )
            if self.frame_count <= 0:
                raise ValueError('frame_count must be positive')
            if len(self.waveform) % self.frame_count != 0:
                raise ValueError('frame_count must be a divisor of len(waveform)')

    @classmethod
    def empty(cls):
        """Clean way of creating empty segments"""
        return cls(None, None, b'', 0, PaStatusFlags.paNoError)

    def is_empty(self) -> bool:
        """Clean way of testing if a segment is empty"""
        cond1 = self.start_date is None
        cond2 = self.end_date is None
        cond3 = len(self.waveform) == 0
        cond4 = self.frame_count == 0
        if cond1 and cond2 and cond3 and cond4:
            return True
        return False

    @property
    def bt(self) -> Optional[Union[int, float]]:
        """Legacy naming: bottom time"""
        return self.start_date

    @property
    def tt(self) -> Optional[Union[int, float]]:
        """Legacy naming: top time"""
        return self.end_date

    @cached_property
    def frame_size(self) -> Optional[int]:
        """Byte count per frame"""
        if self.is_empty():
            return None
        return int(len(self.waveform) / self.frame_count)

    @property
    def duration(self) -> Optional[Union[int, float]]:
        """AudioSegment duration, in the same unit as start and end date"""
        if self.is_empty():
            return None
        return self.end_date - self.start_date

    @property
    def frame_period(self) -> Optional[float]:
        """frame period, in the same unit as start and end date"""
        if self.is_empty():
            return None
        return self.duration / self.frame_count

    @property
    def frame_rate(self) -> Optional[float]:
        """frame rate, in the same unit as start and end date, inversed"""
        if self.is_empty():
            return None
        return 1 / self.frame_period

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}('
            f'start_date={self.start_date}, '
            f'end_date={self.end_date}, '
            f'frame_count={self.frame_count}, '
            f'status_flags={self.status_flags}'
            ')'
        )

    def __lt__(self, other: Union['AudioSegment', int, float]) -> bool:
        """Less than, comparing start_date

        :param other: AudioSegment | int | float
        :return: bool
        """
        self._test_if_not_empty_for_comparison(other)
        other_val = other.start_date if isinstance(other, AudioSegment) else other
        return self.start_date < other_val

    def __gt__(self, other: Union['AudioSegment', int, float]) -> bool:
        """Greater than, comparing start_date

        :param other: AudioSegment | int | float
        :return: bool
        """
        self._test_if_not_empty_for_comparison(other)
        other_val = other.start_date if isinstance(other, AudioSegment) else other
        return self.start_date > other_val

    def __eq__(self, other: 'AudioSegment'):
        """Equal to, comparing start_date and end_date

        :param other: AudioSegment | int | float
        :return: bool
        """
        self._test_if_not_empty_for_comparison(other)
        return self.start_date == other.start_date and self.end_date == other.end_date

    def __getitem__(self, ts: Union[slice, int, float]) -> 'AudioSegment':
        """Slice waveform and return AudioSegment with updated timestamps and data.

        :param ts: timestamp in the form of slice or scalar
        :return: Sliced AudioSegment
        """
        if self.is_empty():
            raise TypeError("can't get items of empty AudioSegment")
        if isinstance(ts, slice):
            return self._get_slice(ts)
        return self._get_frame(ts)

    def __add__(self, other: 'AudioSegment') -> 'AudioSegment':
        """Concatenates two AudioSegments. They must be contiguous."""
        return AudioSegment.concatenate([self, other])

    def __iadd__(self, other: 'AudioSegment') -> 'AudioSegment':
        return self + other

    @staticmethod
    def concatenate(audio_segments: Sequence['AudioSegment']):
        """Join a sequence of AudioSegment. AudioSegments must be contiguous. 
        If all are empty, returns an empty AudioSegment.

        :param audio_datas: list of AudioSegment
        :return: The concatenated AudioSegment.
        """
        audio_segments = {
            audioseg for audioseg in audio_segments if not audioseg.is_empty()
        }
        if not audio_segments:
            return AudioSegment.empty()
        audio_segments = sorted(audio_segments)
        bts, tts = zip(*[(audioseg.bt, audioseg.tt) for audioseg in audio_segments])
        assert all(
            tts[i] == bts[i + 1] for i in range(len(audio_segments) - 1)
        ), 'Audio segments must be contiguous'
        assert (
            len({audioseg.frame_size for audioseg in audio_segments}) == 1
        ), 'Audio segments must all have the same frame_size'
        start_date, end_date = bts[0], tts[-1]
        waveform = b''.join(audioseg.waveform for audioseg in audio_segments)
        frame_count = sum(audioseg.frame_count for audioseg in audio_segments)
        status_flags = PaStatusFlags.paNoError
        for audioseg in audio_segments:
            status_flags |= PaStatusFlags(audioseg.status_flags)

        return AudioSegment(start_date, end_date, waveform, frame_count, status_flags)

    def get_ts_of_frame_index(self, frame_idx: int) -> Tuple[float, float]:
        """get bt and tt of frame at index frame_idx

        :param frame_idx: frame index
        :return: bt, tt
        """
        if not isinstance(frame_idx, int):
            raise TypeError('Index must be an integer')
        if not -self.frame_count <= frame_idx < self.frame_count:
            raise IndexError(
                'Index out of range. '
                f'Must take values in [-{self.frame_count}, {self.frame_count}), '
                f'Got {frame_idx}.'
            )
        frame_idx = frame_idx % self.frame_count
        bt = self.start_date + frame_idx * self.frame_period
        tt = self.start_date + (frame_idx + 1) * self.frame_period
        return bt, tt

    def _get_slice(self, s: slice) -> 'AudioSegment':
        """Slice waveform and return AudioSegment with updated timestamps and data.

        Further info on methodology found in timestamping data stream discussion:
        https://miro.com/app/board/uXjVPsuJtdM=/?share_link_id=872583858418

        Note: slice 'step' is unlikely used in the traditional sense which is how it is implemented.
        A future improvement could use 'step' as a channel selector by filtering out bytes of other
        channels.
        Charlie's EDIT: changed the function so that slice steps are no longer authorized. The
        current class implementation does not support multi-channel audio segments.

        Implemented so that a series of contiguous slices, whatever their ranges and values, will 
        generate a series of AudioSegments that can be concatenated using the concatenate method, 
        i.e. a series of AudioSegments that are either contiguous with no duplicates, or empty.

        If audioseg is an AudioSegment object, then, for any increasing values of t0, t1, ..., tn:
        audioseg[t0:t1] + audioseg[t1:t2] + ... + audioseg[tn_2:tn_1] + audioseg[tn_1:tn] is equal 
        to audioseg[t0:tn]

        More details on the behavior of this method in its corresponding test function in 
        audiostream2py.test.audio_data_test

        NB: If the slice is smaller than a frame period, it is possible this method returns an empty
        AudioSegment, even if the slice is within the AudioSegment time range. This will happen if 
        the slice is located on a single and same frame. This choice has been made to meet the 
        non-duplication and reconcatenation requirements mentionned above.        
        
        :param s: slice object with start and stop by timestamp
        :return: Sliced AudioSegment
        """
        if s.step is not None:
            raise TypeError('Slice steps are not supported.')

        if s.start is not None:
            start_frame = self._nearest_frame_index(ts=s.start, rounding_type='ceil')
        else:
            start_frame = 0

        if s.stop is not None:
            end_frame = self._nearest_frame_index(ts=s.stop, rounding_type='floor')
            if end_frame is not None:
                if (
                    s.stop < self.end_date
                    and (s.stop - self.start_date) % self.frame_period == 0
                ):
                    # Does not include frame i if s.stop == bt of frame i
                    end_frame -= 1
        else:
            end_frame = self.frame_count - 1

        if None in (start_frame, end_frame) or start_frame > end_frame:
            # Happens in invalid slices (descending or out of time range)
            return self.empty()

        start_date, _ = self.get_ts_of_frame_index(start_frame)
        _, end_date = self.get_ts_of_frame_index(end_frame)

        start = start_frame * self.frame_size
        stop = (end_frame + 1) * self.frame_size
        waveform = self.waveform[start:stop]
        frame_count = (end_frame + 1) - start_frame
        return AudioSegment(
            start_date, end_date, waveform, frame_count, self.status_flags
        )

    def _get_frame(self, ts: Union[int, float]) -> 'AudioSegment':
        """Get frame at timestamp.

        :param ts: timestamp microseconds
        :return: Sliced AudioSegment
        """
        if not self.start_date <= ts < self.end_date:
            raise IndexError(
                'Timestamp out of range. '
                f'Must take values in ({self.start_date}, {self.end_date}); '
                f'Got {ts}.'
            )
        frame_idx = self._nearest_frame_index(ts=ts, rounding_type='floor')
        start_date, end_date = self.get_ts_of_frame_index(frame_idx)
        start = frame_idx * self.frame_size
        stop = (frame_idx + 1) * self.frame_size
        waveform = self.waveform[start:stop]
        frame_count = 1

        return AudioSegment(
            start_date, end_date, waveform, frame_count, self.status_flags
        )

    def _nearest_frame_index(
        self, ts, rounding_type: Literal['floor', 'ceil'] = 'ceil'
    ) -> int:
        """Calculate frame index coming just after ('ceil') or just before ('floor') given timestamp

        'ceil' for slice().start, 'floor' for slice().stop
        'floor' for indices

        Further info on methodology found in timestamping data stream discussion:
        https://miro.com/app/board/uXjVPsuJtdM=/?share_link_id=872583858418

        :param ts: timestamp in the same unit as start and end date
        :param rounding_type: 'ceil' or 'floor'
        :return: frame_index or None if ts and rounding_method does not lead to an actual frame
        """
        if rounding_type not in ('floor', 'ceil'):
            raise ValueError(
                f'{type(self).__name__}.nearest_frame_index: rounding_type must be '
                f'literal "floor" or "ceil" but got "{rounding_type}"'
            )
        rounder = getattr(math, rounding_type)
        frame_idx = rounder((ts - self.start_date) / self.frame_period)
        if frame_idx < 0 and rounding_type == 'floor':
            return None
        if frame_idx > self.frame_count - 1 and rounding_type == 'ceil':
            return None
        return min(max(frame_idx, 0), self.frame_count - 1)

    def _test_if_not_empty_for_comparison(self, other) -> None:
        if self.is_empty():
            raise TypeError("Can't compare empty audio segments")
        if isinstance(other, AudioSegment):
            if other.is_empty():
                raise TypeError("Can't compare empty audio segments")



"""Test run PyAudioSourceReader and print info"""
import time

from audiostream2py import PyAudioSourceReader, PaStatusFlags
from audiostream2py.audio import BasePyAudioSourceReader


def _test_run_PyAudioSourceReader(
    readerClass=BasePyAudioSourceReader,
    timestamp_getter=lambda data: data[0],
    in_data_getter=lambda data: data[1],
    status_flags_getter=lambda data: data[4],
    input_device_index=None,
):
    """Run PyAudioSourceReader and print timestamp, byte length, status code, status flag

    :param readerClass: PyAudioSourceReader class or subclass
    :param timestamp_getter: get timestamp from data
    :param in_data_getter: get wf bytes from data
    :param status_flags_getter: get status_flags from data
    """
    from pprint import pprint

    if input_device_index is None:
        print('list_recording_devices')
        pprint(readerClass.list_recording_devices())

    source = readerClass(
        rate=44100,
        width=2,
        channels=1,
        input_device_index=input_device_index,
        frames_per_buffer=4096,
    )
    pprint(source.info)
    try:
        source.open()
        pprint(source.info)
        i = 0
        while i < 600:
            data = source.read()
            if data:
                i += 1
                timestamp = timestamp_getter(data)
                in_data = in_data_getter(data)
                status_flags = status_flags_getter(data)
                print(
                    timestamp,
                    len(in_data),
                    int(status_flags),
                    PaStatusFlags(status_flags),
                )

    finally:
        source.close()
        pprint(source.info)


def _test_run_PyAudioSourceReader_with_StreamBuffer(
    readerClass=PyAudioSourceReader,
    timestamp_getter=lambda data: data.bt,
    in_data_getter=lambda data: data.waveform,
    status_flags_getter=lambda data: data.status_flags,
    input_device_index=None,
    rate=44100,
    maxlen_seconds=10,
    frames_per_buffer=44100,
):
    """Run PyAudioSourceReader and print timestamp, byte length, status code, status flag

    :param readerClass: PyAudioSourceReader class or subclass
    :param timestamp_getter: get timestamp from data
    :param in_data_getter: get wf bytes from data
    :param status_flags_getter: get status_flags from data
    """
    from pprint import pprint

    if input_device_index is None:
        print('list_recording_devices')
        pprint(readerClass.list_recording_devices())

    source = readerClass(
        rate=rate,
        width=2,
        channels=1,
        input_device_index=input_device_index,
        frames_per_buffer=frames_per_buffer,
    )
    pprint(source.info)
    with source.stream_buffer(
        maxlen=source.audio_buffer_size_seconds_to_maxlen(
            maxlen_seconds, rate, frames_per_buffer
        ),
    ) as sb:
        reader = sb.mk_reader()
        time.sleep(6)
        start = reader.head(peek=True).start_date + 0.5 * 1e5
        stop = reader.tail(peek=True).end_date - 0.5 * 1e5
        print(reader[start:stop])
        print(reader[start : start + 0.1 * 1e5])


if __name__ == '__main__':
    # _test_run_PyAudioSourceReader(input_device_index='NexiGo N930AF FHD Webcam Audio')
    # _test_run_PyAudioSourceReader(PyAudioSourceReaderRaiseOnError)

    # _test_run_PyAudioSourceReader(
    #     PyAudioSourceReader,
    #     timestamp_getter=lambda data: data.bt,
    #     in_data_getter=lambda data: data.waveform,
    #     status_flags_getter=lambda data: data.status_flags,
    # )

    _test_run_PyAudioSourceReader_with_StreamBuffer()



"""Tests for AudioSegment"""
import math

from audiostream2py import AudioSegment, PaStatusFlags
from recode import decode_pcm_bytes


def mock_audio_data(rate: int, width: int, channels: int, frames_per_buffer: int):
    def mock_pcm_bytes():
        """Value of sample is the frame index"""
        return b''.join(
            i.to_bytes(width, 'little') * channels for i in range(frames_per_buffer)
        )

    return AudioSegment(
        start_date=0,
        end_date=1e6 * frames_per_buffer / rate,
        waveform=mock_pcm_bytes(),
        frame_count=frames_per_buffer,
        status_flags=PaStatusFlags.paNoError,
    )


def test_process_to_find_first_sample_of_session():
    def first_sample_of_session(bt_i, t_start, tt_i, n):
        """Process defined in timestamping discussion"""
        m = math.ceil((t_start - bt_i) / (tt_i - bt_i) * n)
        t_m = bt_i + m / n * (tt_i - bt_i)
        return m, t_m

    ad = mock_audio_data(44100, 2, 1, 4096)
    t_start = 0.05 * 1e6
    s_start, s_time = first_sample_of_session(
        ad.start_date, t_start, ad.end_date, ad.frame_count
    )
    session = ad[t_start:]
    session_start = decode_pcm_bytes(session.waveform, 2, 1)[0]
    assert s_start == session_start, 'Wrong Start Sample'
    assert s_time == session.start_date, 'Wrong Start Time'


def test_process_to_find_last_sample_of_session():
    def last_sample_of_session(bt_i, t_stop, tt_i, n):
        """Process defined in timestamping discussion"""
        m = math.floor((t_stop - bt_i) / (tt_i - bt_i) * n)
        t_m1 = bt_i + (m + 1) / n * (tt_i - bt_i)
        return m, t_m1

    ad = mock_audio_data(44100, 2, 1, 4096)
    t_stop = 0.05 * 1e6
    s_stop, s_time = last_sample_of_session(
        ad.start_date, t_stop, ad.end_date, ad.frame_count
    )
    session = ad[:t_stop]
    session_stop = decode_pcm_bytes(session.waveform, 2, 1)[-1]
    assert s_stop == session_stop, 'Wrong End Sample'
    assert s_time == session.end_date, 'Wrong End Time'


def _create_test_audiosegment() -> AudioSegment:
    start_date = 10
    end_date = 20
    waveform = bytes(range(15))
    frame_count = 5
    status_flags = PaStatusFlags.paNoError
    return AudioSegment(start_date, end_date, waveform, frame_count, status_flags)


def test_properties() -> None:
    audioseg = _create_test_audiosegment()
    _assert_equality(audioseg.bt, 10)
    _assert_equality(audioseg.tt, 20)
    _assert_equality(audioseg.frame_size, 3)
    _assert_equality(audioseg.duration, 10)
    _assert_equality(audioseg.frame_period, 2)
    _assert_equality(audioseg.frame_rate, 0.5)


def test_comparisons() -> None:
    audioseg = _create_test_audiosegment()
    assert 9 < audioseg < 11
    assert (
        AudioSegment(9, 10, b'\xff', 1, None)
        < audioseg
        < AudioSegment(11, 12, b'\xff', 1, None)
    )
    assert audioseg == AudioSegment(10, 20, b'\xff', 1, None)


def test_empty() -> None:
    assert AudioSegment.empty().is_empty()


def test_get_slice() -> None:
    audioseg = _create_test_audiosegment()
    _assert_is_empty(audioseg[7:9])
    _assert_is_empty(audioseg[7:10])
    _assert_audiosegment_values(audioseg[7:10.2], 10, 12, bytes(range(0, 3)), 1)
    _assert_audiosegment_values(audioseg[10:10.2], 10, 12, bytes(range(0, 3)), 1)
    _assert_is_empty(audioseg[10.1:10.2])
    _assert_audiosegment_values(audioseg[10.1:14], 12, 14, bytes(range(3, 6)), 1)
    _assert_is_empty(audioseg[14:14])
    _assert_audiosegment_values(audioseg[14:18], 14, 18, bytes(range(6, 12)), 2)
    _assert_audiosegment_values(audioseg[14:20], 14, 20, bytes(range(6, 15)), 3)
    _assert_audiosegment_values(audioseg[14:20.2], 14, 20, bytes(range(6, 15)), 3)
    _assert_is_empty(audioseg[18.1:20.2])
    _assert_is_empty(audioseg[20:20.2])
    _assert_is_empty(audioseg[20.1:20.2])


def test_get_frame() -> None:
    audioseg = _create_test_audiosegment()
    _assert_audiosegment_values(audioseg[10], 10, 12, bytes(range(0, 3)), 1)
    _assert_audiosegment_values(audioseg[10.1], 10, 12, bytes(range(0, 3)), 1)
    _assert_audiosegment_values(audioseg[12], 12, 14, bytes(range(3, 6)), 1)
    _assert_audiosegment_values(audioseg[18], 18, 20, bytes(range(12, 15)), 1)
    _assert_audiosegment_values(audioseg[18.1], 18, 20, bytes(range(12, 15)), 1)


def test_get_ts_of_frame_index() -> None:
    audioseg = _create_test_audiosegment()
    _assert_equality(audioseg.get_ts_of_frame_index(0), (10, 12))
    _assert_equality(audioseg.get_ts_of_frame_index(1), (12, 14))
    _assert_equality(audioseg.get_ts_of_frame_index(4), (18, 20))
    _assert_equality(audioseg.get_ts_of_frame_index(-1), (18, 20))
    _assert_equality(audioseg.get_ts_of_frame_index(-5), (10, 12))


def test_concatenate() -> None:
    audioseg1 = AudioSegment(-1.5, 0, bytes(range(0, 15)), 3, PaStatusFlags.paNoError)
    audioseg2 = AudioSegment(
        0, 3, bytes(range(15, 45)), 6, PaStatusFlags.paInputOverflow
    )
    audioseg3 = AudioSegment(
        3, 4.5, bytes(range(45, 60)), 3, PaStatusFlags.paInputUnderflow
    )
    empty = AudioSegment.empty()
    concat = AudioSegment.concatenate(
        [audioseg1, audioseg3, empty, audioseg2, audioseg1]
    )
    _assert_audiosegment_values(concat, -1.5, 4.5, bytes(range(0, 60)), 12)
    assert concat.status_flags == 3


def _assert_equality(item1, item2) -> None:
    assert item1 == item2, f'Expected {item2}; Got {item1}'


def _assert_audiosegment_values(
    audioseg: AudioSegment, bt, tt, bytes_, frame_count
) -> None:
    _assert_equality(audioseg.start_date, bt)
    _assert_equality(audioseg.end_date, tt)
    _assert_equality(audioseg.waveform, bytes_)
    _assert_equality(audioseg.frame_count, frame_count)


def _assert_is_empty(audioseg: AudioSegment) -> None:
    assert audioseg.is_empty(), f'Expected empty; Got {audioseg}'



"""BufferReader specific to AudioSegment"""

from typing import List

from stream2py import BufferReader

from audiostream2py.data import AudioSegment


class OverlapsPastError(IndexError):
    pass


class OverlapsFutureError(IndexError):
    pass


class AudioBufferReader(BufferReader):
    def __getitem__(self, item) -> AudioSegment:
        if isinstance(item, slice):
            return self.get_slice(item)
        with self._buffer.reader_lock() as reader:
            return reader.find_le(item)

    def get_slice(self, s: slice) -> AudioSegment:
        """Slice starting from start_date and stopping at end_date. Trims first and last
        AudioSegment based on timestamps. Then joins all the AudioSegment objects.

        :param s: slice object
        :return: concatenated AudioSegment
        """
        with self._buffer.reader_lock() as reader:
            if s.start:
                head: AudioSegment = reader[0]
                if s.start < head.start_date:
                    raise OverlapsPastError(
                        f'You asked for {s}, but the buffer starts from: {head.start_date}'
                    )
                start = reader.key(reader.find_le(s.start))
            else:
                start = None

            if s.stop:
                tail: AudioSegment = reader[-1]
                if s.stop > tail.end_date:
                    raise OverlapsFutureError(
                        f'You asked for {s}, but the buffer stops at: {tail.end_date}'
                    )
                if tail.start_date < s.stop < tail.end_date:
                    stop = reader.key(tail)
                else:
                    stop = reader.key(reader.find_le(s.stop))
            else:
                stop = None

            items: List[AudioSegment] = reader.range(start, stop, s.step)
        if s.start:
            items[0] = items[0][s.start :]
        if s.stop:
            items[-1] = items[-1][: s.stop]
        return AudioSegment.concatenate(items)



"""PyAudioSourceReader Status Flag Enums"""

from enum import IntFlag

import pyaudio


class PaStatusFlags(IntFlag):
    """Enum to check status_flag for error codes.

    >>> from audiostream2py.audio import PaStatusFlags
    >>> PaStatusFlags(0)
    <PaStatusFlags.paNoError: 0>
    >>> PaStatusFlags(2)
    <PaStatusFlags.paInputOverflow: 2>
    >>> PaStatusFlags(3)
    <PaStatusFlags.paInputOverflow|paInputUnderflow: 3>
    >>> PaStatusFlags.paInputOverflow in PaStatusFlags(3)  # Check if contains certain error
    True
    >>> PaStatusFlags.paNoError == PaStatusFlags(3)  # Check for no error
    False
    """

    paNoError = pyaudio.paNoError  # 0
    paInputUnderflow = pyaudio.paInputUnderflow  # 1
    paInputOverflow = pyaudio.paInputOverflow  # 2
    paOutputUnderflow = pyaudio.paOutputUnderflow  # 4
    paOutputOverflow = pyaudio.paOutputOverflow  # 8
    paPrimingOutput = pyaudio.paPrimingOutput  # 16
    hostTimeSync = 32  # Timestamp adjusted to avoid overlap due to time sync with NTP


class PaCallbackReturnCodes(IntFlag):
    """Enum of valid _stream_callback return codes.
    Only used by PyAudioSourceReader._stream_callback"""

    paContinue = pyaudio.paContinue
    paComplete = pyaudio.paComplete
    paAbort = pyaudio.paAbort



"""Utils for testing

Generate waveforms with ``inverted_tiles_wf`` or chunks thereof with `inverted_tiles`.

"""


def inverted_tiles_wf(motif=(1, 2, 3, 4, 5), cap=20000, wf_length=21 * 2048):
    """Get a (possibly infinite) waveform through inverted_tiles chunks

    These waveforms are originally meant to be used with [1,...,n] motifs and used for
    diagnosis. Indeed, the samples alternate sign (so waveform can be made audible),
    but checking for the absolute value of the samples, you should see incrementing
    integers.

    :param motif: The iterable to use as the (incremented and inverted) pattern to
        make waveforms out of.
    :param cap: The maximum absolute value to include in a waveform.
        Also determines when we resset the chk to be the motif.
    :param wf_length: Length of the desired waveform, or None for an infinite iterable

    >>> wf = inverted_tiles_wf()
    >>> len(wf)
    43008
    >>> wf[:15]
    [1, 2, 3, 4, 5, -6, -7, -8, -9, -10, 11, 12, 13, 14, 15]
    >>> inverted_tiles_wf(motif=(1, 2, 3), cap=10, wf_length=15)
    [1, 2, 3, -4, -5, -6, 7, 8, 9, -10, -11, -12, 1, 2, 3]

    """
    from itertools import chain, islice

    chks = inverted_tiles(motif=motif, cap=cap)
    it = chain.from_iterable(chks)
    if wf_length is not None:
        return list(islice(it, wf_length))
    else:
        return it


def inverted_tiles(motif=(1, 2, 3, 4, 5), cap=20000):
    """Infinite generator of chunks taken by transforming the input motif.
    The iterative transformation is constituted of a translation by the length of the
    motif and an inversion of the sign.

    :param motif: The iterable to use as the (incremented and inverted) pattern to
        make waveforms out of.
    :param cap: The maximum absolute value to include in a waveform.
        Also determines when we resset the chk to be the motif.

    >>> g = inverted_tiles()
    >>> next(g), next(g), next(g)
    ([1, 2, 3, 4, 5], [-6, -7, -8, -9, -10], [11, 12, 13, 14, 15])
    >>> g = inverted_tiles(motif=(1, 2, 3), cap=10)
    >>> [*next(g), *next(g), *next(g), *next(g)]
    [1, 2, 3, -4, -5, -6, 7, 8, 9, -10, -11, -12]

    """
    # input preprocessing
    if isinstance(motif, int):
        motif = range(1, motif + 1)
    motif = list(motif)
    current_chk = motif
    motif_length = len(current_chk)
    sign = 1
    while True:
        yield current_chk
        if max(abs(x) for x in current_chk) >= cap:  # if we're above the cap
            current_chk = motif  # reset the current_motif
        else:
            sign *= -1  # invert sign
            # and make the next current_chk be
            # the sign-inverted motif_length-incremented version of the current one
            current_chk = [sign * (abs(x) + motif_length) for x in current_chk]



"""
PyAudio Source
--------------
This source requires PyAudio.
`Find PyAudio installation instructions here <https://people.csail.mit.edu/hubert/pyaudio/>`_.
::
    pip install PyAudio

.. autoclass:: audiostream2py.audio.PyAudioSourceReader()
    :members:
    :show-inheritance:

    .. automethod:: __init__

.. autoclass:: audiostream2py.audio.PaStatusFlags
    :members:
    :undoc-members:
    :show-inheritance:
    :member-order: bysource


PyAudio Mixins
--------------
Example of how to use mixins to manipulate behavior to suit your needs.

.. autoclass:: audiostream2py.audio.FillErrorWithZeroesMixin
    :members:

.. autoclass:: audiostream2py.audio.DictDataMixin
    :members:

.. autoclass:: audiostream2py.audio.RaiseRuntimeErrorOnStatusFlagMixin
    :members:

Example PyAudio Mixin Usage
---------------------------
Mixins must be subclassed in order of precedence from left to right.
Meaning the base class at the right and mixins with overloading methods to the left.

.. autoclass:: audiostream2py.audio.PyAudioSourceReaderRaiseOnError
    :members:
    :show-inheritance:

.. autoclass:: audiostream2py.audio.PyAudioSourceReaderRaiseOnError
    :members:
    :show-inheritance:

.. autoclass:: audiostream2py.audio.PyAudioSourceReaderWithZeroedErrorsAndDictData
    :members:
    :show-inheritance:
"""

__all__ = [
    'PyAudioSourceReader',
    'PaStatusFlags',
    'PaCallbackReturnCodes',
    'get_input_device_index',
    'AudioSegment',
]

from collections import deque
from contextlib import suppress, contextmanager
import math
from typing import Generator, List, Callable, Union
import re

import pyaudio
from stream2py import SourceReader
from stream2py.utility.typing_hints import ComparableType

from audiostream2py.reader import AudioBufferReader
from audiostream2py.data import AudioSegment
from audiostream2py.enum import PaCallbackReturnCodes, PaStatusFlags


def list_recording_device_index_names():
    """List (index, name) of available recording devices. Returns a list of the form:
    [(2, 'MacBook Pro Microphone'), (4, 'BluetoothHeadset'), (6, 'ZoomAudioDevice')]
    """
    return sorted(
        (d['index'], d['name'])
        for d in PyAudioSourceReader.list_device_info()
        if d['maxInputChannels'] > 0
    )


def _list_device_info():
    return PyAudioSourceReader.list_device_info()


def _match_device_info(filt: Callable):
    return filter(filt, _list_device_info())


def match_device_info(filt: Callable, assert_unique=True):
    """Find a device ('s info) through a function of its info (dict)"""
    it = _match_device_info(filt)
    info = next(it, None)
    if info is None:
        raise RuntimeError('No device infos matched your filter')
    if assert_unique:
        if next(it, None) is not None:
            raise RuntimeError('More than one device info dict matched your filter')
    return info


def is_recording_device(device_info: dict):
    """Says if a device (info) is a recording device"""
    return device_info.get('maxInputChannels', 0) > 0


def get_recording_device_info_by_name(name_pattern, assert_unique=True):
    """Find a recording device by matching name pattern and return info dict"""
    name_pattern = re.compile(name_pattern)

    def filt(device_info):
        return name_pattern.search(device_info['name']) and is_recording_device(
            device_info
        )

    return match_device_info(filt, assert_unique)


def get_input_device_index(input_device=None, input_device_index=None, verbose=True):
    info = None
    # You can't specify both input_device_index and input_device!
    if input_device is not None and input_device_index is not None:
        raise RuntimeError(
            f"You can't have specify both input_device and input_device_index!: "
            f'{input_device=} and {input_device_index=}'
        )
    # If no information on the device is given, ask pyaudio to find a default
    if input_device is None and input_device_index is None:
        info = PyAudioSourceReader.get_default_input_device_info()
        if verbose:
            import json

            print(
                f"Will use {info['name']} (index={info['index']}) as an input device",
                end='\r',
            )
            print(f"It's info:\n{json.dumps(info, indent=2)}", end='\r')
    else:
        input_device = input_device or input_device_index
        if isinstance(input_device, int):
            info = {'index': input_device}
        elif isinstance(input_device, Callable):
            info = match_device_info(filt=input_device)
        else:  # should be string or re.compile instance
            info = get_recording_device_info_by_name(name_pattern=input_device)
    return info['index']


class BasePyAudioSourceReader(SourceReader):
    _pyaudio_instance = None

    def __init__(
        self,
        *,
        rate=44100,
        width=2,
        unsigned=True,
        channels=1,
        input_device_index=None,
        frames_per_buffer=1024,
        input_device=None,
        verbose=True,
    ):
        """

        :param rate: Specifies the desired sample rate (in Hz)
        :param width: Sample width in bytes (1, 2, 3, or 4)
        :param unsigned: For 1 byte width, specifies signed or unsigned format.
            Ignored if byte width is not 1.
        :param channels: The desired number of input channels.
            Ignored if input_device is not specified (or None).
        :param input_device: Specification of what input device to use.
        :param input_device_index: Legacy specification of input Device to use.
            Has to be the integer index, unlike input_device which can be a name too.
            Unspecified (or None) uses default device.
        :param frames_per_buffer: Specifies the number of frames per buffer.
        :param verbose: Permission to print stuff when we feel like it?

        """
        super().__init__()
        self._init_kwargs = {
            k: v for k, v in locals().items() if k not in ('self', '__class__')
        }
        input_device_index = get_input_device_index(
            input_device=input_device,
            input_device_index=input_device_index,
            verbose=verbose,
        )
        with self._pyaudio() as pa:
            input_format = pa.get_format_from_width(width, unsigned)
            pa.is_format_supported(
                rate=rate,
                input_device=input_device_index,
                input_channels=channels,
                input_format=input_format,
                output_device=None,
                output_channels=None,
                output_format=None,
            )

        self._pyaudio_open_params = {
            'rate': rate,
            'channels': channels,
            'format': input_format,
            'input': True,
            'output': False,
            'input_device_index': input_device_index,
            'output_device_index': None,
            'frames_per_buffer': frames_per_buffer,
            'start': True,
            'input_host_api_specific_stream_info': None,
            'output_host_api_specific_stream_info': None,
            'stream_callback': self._stream_callback,
        }

        self._fp = None
        self._open_time = None
        self.data = deque()
        self.buffer_start = None
        self.buffer_end = None
        self._init_vars()

    def _init_vars(self):
        if self._fp:
            self.close()
        self._fp = None
        if self.data:
            self.data.clear()

        self.buffer_start = None
        self.buffer_end = None

    def __repr__(self):
        def quote_strings(x):
            return f"'{x}'" if isinstance(x, str) else x

        args_string = ', '.join(
            f'{k}={quote_strings(v)}' for k, v in self._init_kwargs.items()
        )
        return f'{type(self).__name__}({args_string})'

    @classmethod
    def get_default_input_device_info(cls):
        with cls._pyaudio() as p:
            return p.get_default_input_device_info()

    @property
    def sleep_time_on_read_none_s(self) -> float:
        """One third of the expected rate frames_per_buffer will be filled"""
        seconds_per_read = (
            self._init_kwargs['frames_per_buffer'] / self._init_kwargs['rate']
        )
        one_third_read_rate = seconds_per_read / 3
        return one_third_read_rate

    @property
    def sr(self) -> Union[int, float]:
        return self._init_kwargs['rate']

    @property
    def info(self) -> dict:
        """
        Provides a dict with init kwargs, bt, and device info

        >>> from audiostream2py.audio import PyAudioSourceReader
        >>> from pprint import pprint
        >>> source = PyAudioSourceReader(
        ... rate=44100, width=2, channels=1,
        ... input_device_index=7, frames_per_buffer=4096)  # doctest: +SKIP
        >>> source.open()  # doctest: +SKIP
        >>> pprint(source.info) # doctest: +SKIP
        {'bt': 1582851038965183,
         'channels': 1,
         'device_info': {'defaultHighInputLatency': 0.021333333333333333,
                         'defaultHighOutputLatency': 0.021333333333333333,
                         'defaultLowInputLatency': 0.021333333333333333,
                         'defaultLowOutputLatency': 0.021333333333333333,
                         'defaultSampleRate': 48000.0,
                         'hostApi': 0,
                         'index': 7,
                         'maxInputChannels': 128,
                         'maxOutputChannels': 128,
                         'name': 'sysdefault',
                         'structVersion': 2},
         'frames_per_buffer': 4096,
         'input_device_index': 7,
         'rate': 44100,
         'unsigned': True,
         'width': 2}
        >>> source.close()  # doctest: +SKIP

        :return: dict
        """
        _info = {'start_date': self._open_time}
        _info.update(**self._init_kwargs)
        with suppress(Exception):
            _info.update(
                device_info=self._pyaudio_instance.get_device_info_by_index(
                    self._init_kwargs['input_device_index']
                )
            )
        return _info

    def key(self, data) -> ComparableType:
        """
        :param data: (start_date, end_date, waveform, frame_count, time_info, status_flags)
        :return: timestamp
        """
        return data[1]

    def data_to_append(
        self, start_date, end_date, waveform, frame_count, time_info, status_flags
    ):
        """Overloaded data_to_append method change the shape of read outputs of PyAudioSourceReader.
        The key method must also be overloaded to return timestamp from the new shape.

        :param start_date: bt timestamp
        :param end_date: tt timestamp
        :param waveform: recorded input frames in bytes
        :param frame_count: number of frames, sample count or waveform
        :param time_info: http://portaudio.com/docs/v19-doxydocs/structPaStreamCallbackTimeInfo.html
        :param status_flags: PaStatusFlags
        :return: (start_date, end_data, waveform, frame_count, time_info, status_flags)
        """
        return start_date, end_date, waveform, frame_count, time_info, status_flags

    def open(self):
        """Will first close if already open and clear any data before starting the audio stream"""
        self._init_vars()
        self._init_pyaudio()
        self._open_time = self.get_timestamp()
        self._fp = self._pyaudio_instance.open(**self._pyaudio_open_params)

    def close(self):
        """Stop audio stream"""
        with suppress(Exception):
            self._fp.stop_stream()
        with suppress(Exception):
            self._fp.close()
        self._terminate_pyaudio()

    def read(self):
        """Returns one data item as structured by PyAudioSourceReader.data_to_append

        :return: (timestamp, waveform, frame_count, time_info, status_flags)
        """
        if len(self.data):
            return self.data.popleft()
        return None

    def _stream_callback(self, in_data, frame_count, time_info, status_flags):
        """Get buffer end timestamp based on the system time when data is sent to callback. Buffer
        start timestamp is the buffer end of the last buffer. The very first start timestamp is
        calculated back with frame count and sample rate.
        See _stream_callback in https://people.csail.mit.edu/hubert/pyaudio/docs/#class-stream

        :param in_data: recorded input data, waveform
        :param frame_count: number of frames, sample count
        :param time_info: dictionary
        :param status_flags: PaStatusFlags
        :return: None, PaCallbackReturnCodes.paContinue
        """
        _additional_status_flags = self._set_buffer_start_end(frame_count)
        status_flags = PaStatusFlags(status_flags) | _additional_status_flags
        self.data.append(
            self.data_to_append(
                self.buffer_start,
                self.buffer_end,
                in_data,
                frame_count,
                time_info,
                status_flags,
            )
        )
        return None, PaCallbackReturnCodes.paContinue

    def _set_buffer_start_end(self, frame_count):
        """Set buffer end timestamp based on the system time when data is sent to callback. Buffer
        start timestamp is the buffer end of the last buffer. The very first start timestamp is
        calculated back with frame count and sample rate.

        :param frame_count: used to estimate start time of very first buffer
        :return: status flag
        """
        status_flag = PaStatusFlags.paNoError
        buffer_end_timestamp = self.get_timestamp()

        if self.buffer_end is None:
            t_len = frame_count * self.timestamp_seconds_to_unit_conversion / self.sr
            self.buffer_start = buffer_end_timestamp - t_len
        else:
            if buffer_end_timestamp < (
                _adjustment := self.buffer_end + frame_count / 1e6
            ):
                buffer_end_timestamp = _adjustment
                status_flag = PaStatusFlags.hostTimeSync
            self.buffer_start = self.buffer_end
        self.buffer_end = buffer_end_timestamp
        return status_flag

    @classmethod
    def _init_pyaudio(cls) -> pyaudio.PyAudio:
        cls._terminate_pyaudio()
        cls._pyaudio_instance = pyaudio.PyAudio()
        return cls._pyaudio_instance

    @classmethod
    def _terminate_pyaudio(cls):
        with suppress(Exception):
            if cls._pyaudio_instance is not None and isinstance(
                cls._pyaudio_instance, (pyaudio.PyAudio,)
            ):
                cls._pyaudio_instance.terminate()
                cls._pyaudio_instance = None

    @classmethod
    @contextmanager
    def _pyaudio(cls) -> Generator[pyaudio.PyAudio, None, None]:
        try:
            yield cls._init_pyaudio()
        finally:
            cls._terminate_pyaudio()

    @classmethod
    def list_device_info(cls) -> List[dict]:
        """
        .. todo::
            * filter for only devices with input channels

        :return: list
        """
        with cls._pyaudio() as pa:
            return [
                pa.get_device_info_by_index(idx) for idx in range(pa.get_device_count())
            ]

    @staticmethod
    def audio_buffer_size_seconds_to_maxlen(
        buffer_size_seconds, rate, frames_per_buffer
    ) -> int:
        """Calculate maxlen for StreamBuffer to keep a minimum of buffer_size_seconds of data on
        buffer

        :param buffer_size_seconds: desired length of StreamBuffer in seconds
        :param rate: sample rate
        :param frames_per_buffer: number of frames per buffer
        :return: maxlen for StreamBuffer
        """
        seconds_per_read = frames_per_buffer / rate
        return math.ceil(buffer_size_seconds / seconds_per_read)

    @classmethod
    def list_recording_devices(cls):
        """List names of available recording devices"""
        return sorted(
            d['name'] for d in cls.list_device_info() if d['maxInputChannels'] > 0
        )

    @classmethod
    def info_of_input_device_index(cls, input_device_index):
        return next(
            (
                dev
                for dev in cls.list_device_info()
                if dev['index'] == input_device_index
            ),
            None,
        )


class FillErrorWithOnesMixin:
    """Mixin to handle all pyaudio's error status flags by filling the error time gap with zeros"""

    FILL_VALUE = 1
    _first_error_timestamp = None
    _error_status_flag = None

    def _stream_callback(self, in_data, frame_count, time_info, status_flags):
        """On paInputOverflow status flag error code, create wf bytes of value one for the entire
        duration of the error to replace garbled data in addition to the base behavior.

        :param in_data: recorded input data, waveform
        :param frame_count: number of frames, sample count
        :param time_info: dictionary
        :param status_flags: PaStatusFlags
        :return: None, PaCallbackReturnCodes.paContinue
        """

        if PaStatusFlags(status_flags) == PaStatusFlags.paInputOverflow:
            self.buffer_start = None
            if self._first_error_timestamp is None:
                # track when errors started
                self._first_error_timestamp = self.buffer_end
                # recalculate next timestamps in _set_buffer_start_end
                self.buffer_end = None
                # track what errors occurred
                self._error_status_flag = PaStatusFlags(status_flags)
            else:
                # use OR to mark any new error status flags
                self._error_status_flag |= PaStatusFlags(status_flags)
        else:
            _additional_status_flags = self._set_buffer_start_end(frame_count)
            status_flags = PaStatusFlags(status_flags) | _additional_status_flags

            if self._first_error_timestamp is not None:
                # first ok status after there was an error status
                fill_data, fill_count = self._fill_time_interval_with_ones(
                    self._first_error_timestamp, self.buffer_start
                )
                self.data.append(
                    self.data_to_append(
                        self._first_error_timestamp,
                        self.buffer_start,
                        fill_data,
                        fill_count,
                        {},
                        self._error_status_flag | status_flags,
                    )
                )
                self._first_error_timestamp = None
                self._error_status_flag = None

            self.data.append(
                self.data_to_append(
                    self.buffer_start,
                    self.buffer_end,
                    in_data,
                    frame_count,
                    time_info,
                    status_flags,
                )
            )
        return None, PaCallbackReturnCodes.paContinue

    def _fill_time_interval_with_ones(self, first_error_status_ts, first_ok_status_ts):
        """Create wf bytes of value zero for the entire duration of the error to replace garbled
        data

        :param first_error_status_ts: Error bt
        :param first_ok_status_ts: Error tt
        :return: in_data, frame_count
        """

        single_fill_sample = (
            self.FILL_VALUE.to_bytes(self._init_kwargs['width'], 'little')
            * self._init_kwargs['channels']
        )
        samples_per_time_unit = self.sr / self.timestamp_seconds_to_unit_conversion

        delta_time = first_ok_status_ts - first_error_status_ts
        sample_count = int(delta_time * samples_per_time_unit)
        wf_bytes = single_fill_sample * sample_count
        return wf_bytes, sample_count


class DictDataMixin:
    """Mixin to reduce data to a dict with bt, tt, wf, and status_flag. Removing typically discarded
    information.
    """

    def data_to_append(
        self, start_date, end_data, waveform, frame_count, time_info, status_flags
    ):  # pylint: disable=W0613
        """Simplify data only

        :param timestamp: start time of waveform
        :param waveform: recorded input data
        :param frame_count: discarded
        :param time_info: discarded
        :param status_flags: PaStatusFlags error codes
        :return: {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}
        """
        return {
            'bt': start_date,
            'tt': end_data,
            'wf': waveform,
            'status_flags': status_flags,
        }

    def key(self, data) -> ComparableType:
        """
        :param data: {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}
        :return: data['tt']
        """
        return data['tt']


class AudioSegmentMixin:
    """Mixin to put data into AudioSegment"""

    buffer_reader_class = AudioBufferReader  # BufferReader specific to AudioSegment

    def data_to_append(
        self,
        start_date,
        end_data,
        waveform,
        frame_count,
        time_info,  # pylint: disable=W0613
        status_flags,
    ) -> AudioSegment:
        """Puts data into AudioSegment

        :param start_date: starting timestamp
        :param end_data: ending timestamp
        :param waveform: recorded input data
        :param frame_count: frame count of current buffer
        :param time_info: discarded
        :param status_flags: PaStatusFlags error codes
        :return: {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}
        """
        return AudioSegment(start_date, end_data, waveform, frame_count, status_flags)

    def key(self, data: AudioSegment) -> ComparableType:
        """AudioSegment is a ComparableType

        :param data: {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}
        :return: AudioSegment
        """
        return data.start_date


class PyAudioSourceReader(
    AudioSegmentMixin, FillErrorWithOnesMixin, BasePyAudioSourceReader
):
    """PyAudioSourceReader changed to handle errors and serve data in an easy-to-read dataclass."""
